/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : VRPGAOperation.h                                         */
/*                                                                       */
/*       Date : June  16, 1999                                           */
/* Updated at : June  16, 1999                                           */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*  
    Mutation Operator for Airline Crew Scheduling            
*/
/*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <fstream.h>

#include <deffunc.h>
#include <MutationOperator.h>
#include <RandomGenerator.h>

#include "VRPGAOperation.h"
#include "VRPChromosomes.h"


/*************************************************************************/
/*************************************************************************/
VRPGAOperation::find1PX( EdgeListPtr edLst1, int fIndex, 
			 EdgeListPtr edLst2, int sIndex, int& rfIndex, 
			 int& rsIndex, EdgeListPtr* ret, VRProblem* prob) {

   int retValue = 0;
   EdgeGraph *edgeGraph = prob->getEdgeGraph();
   EdgeListPtr tmp1;
   EdgeListPtr tmp2;
   EdgePtr    edge1;
   EdgePtr    edge2;
   vrpcapacity v1=0;
   vrpcapacity v2=0;
   vrpcapacity vt1=0;
   vrpcapacity vt2=0;
   vrpcapacity MVCap=prob->getMaximumCapacity();
   Customer cus;
   int i, j, m, n, fr1, to1, fr2, to2, k, l;

   //if(myDebug) 
   cout << " In VRPGAOperation::find1PX() " << endl;
   //if(myDebug) cout << " edLst1 " << *edLst1 << endl;
   //if(myDebug) cout << " edLst2 " << *edLst2 << endl;

   v1=0; vt1=edLst1->getDemand(); m=fIndex;
   v2=0; vt2=edLst2->getDemand(); 

   v1 = edLst1->getDemandTo(fIndex+1,prob);
   //if(myDebug) cout << " v1 " << v1 << endl;

   while(m<(edLst1->getNumberOfEdges()-1) && !retValue) {
     edge1 = edLst1->getNthEdge(m);
     //if(myDebug) cout << " edge1 " << *edge1 ;
     if(edge1!=NULL) { fr1=edge1->getStartNode(); to1=edge1->getEndNode(); }
     n=sIndex; v2 = edLst2->getDemandTo(n+1,prob);
     while(edge1!=NULL && n<(edLst2->getNumberOfEdges()-1) && !retValue ) {
       edge2 = edLst2->getNthEdge(n);
       if(edge2!=NULL) {
          //if(myDebug) cout << " edge2 " << *edge2 << endl;
	  fr2 = edge2->getStartNode();
	  to2 = edge2->getEndNode();

	  retValue = ( fr1!=fr2 || to1!=to2 ) && 
	     (v1+vt2-v2)<=MVCap && (v2+vt1-v1)<=MVCap && 
	     edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
	     (edge1->getEndDateFirstNode()+edgeGraph->getTravelTime(fr1,to2))<=edge2->getStartDateSecondNode() &&
	     //edgeGraph->canConnectWithoutShift(fr1,to2,edge1->getStartDate(),edge2->getStartDateSecondNode()) &&
             edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) &&
	     (edge2->getEndDateFirstNode()+edgeGraph->getTravelTime(fr2,to1))<=edge1->getStartDateSecondNode() ;
             //edgeGraph->canConnectWithoutShift(fr2,to1,edge2->getStartDate(),edge1->getStartDateSecondNode()) ;

          if(retValue) {
	     tmp1 = defineELists(edLst2,n,edLst1,m,prob);
             tmp2 = defineELists(edLst1,m,edLst2,n,prob);
             retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
	     if(!retValue) { delete tmp1; delete tmp2; tmp1=tmp2=NULL; }

	  }// end of if

	  if(prob->getCustomer(to2,cus)) v2+=cus.getDemand();

       }// end of if

       if(!retValue) n++;
	      
     }// end of while

     if(!retValue) m++;
     if(prob->getCustomer(to1,cus)) v1+=cus.getDemand();

   }// end of while

   if(retValue) { ret[0]=tmp1; ret[1]=tmp2; }
   rfIndex=m; rsIndex=n; 

   //if(myDebug) 
   cout << " Out of VRPGAOperation::find1PX() " << endl;

   return retValue;

}// end of find1PX()

/*************************************************************************/
/*************************************************************************/
VRPGAOperation::find1PX( EdgeListPtr edLst1, int fIndex, 
			 EdgeListPtr edLst2, int sIndex,
	                 int& rfIndex, int& rsIndex, VRProblem* prob) {
 EdgeListPtr ret[2];
 int retValue = find1PX(edLst1,fIndex,edLst2,sIndex,rfIndex,rsIndex,ret,prob);
 if(retValue) { delete ret[0]; delete ret[1]; }
 return retValue;
}// end of find1PX()

/*************************************************************************/
/*************************************************************************/
VRPGAOperation::find1PXG( EdgeListPtr edLst1, int fIndex, 
			  EdgeListPtr edLst2, int sIndex, 
			  int& rfIndex, int& rsIndex, 
			  EdgeListPtr* ret, int greedy, 
			  VRProblem* prob, int fakeGreedy) {

   int retValue = 0;
   EdgeGraph *edgeGraph = prob->getEdgeGraph();
   EdgeListPtr tmp1;
   EdgeListPtr tmp2;
   EdgePtr    edge1;
   EdgePtr    edge2;
   vrpcapacity v1=0;
   vrpcapacity v2=0;
   vrpcapacity vt1=0;
   vrpcapacity vt2=0;
   vrpcapacity MVCap=prob->getMaximumCapacity();
   Customer cus;
   int i, j, m, n, fr1, to1, fr2, to2, k, l, fm, fn, found=0, stay=1;
   EdgeListPtr retE[2];
   float ccost, cc1,cc2, totdemp, totdemc, pdcostp, pdcostc;

   //if(myDebug) 
   cout << " In VRPGAOperation::find1PXG(EdgeListPtr) " << endl;
   if(!edLst1->isValid(prob,1)) { edLst1->printTabular(cout); exit(1); }
   if(!edLst2->isValid(prob,1)) { edLst2->printTabular(cout); exit(1); }
   //if(myDebug) cout << " edLst1 " << *edLst1 << endl;
   //if(myDebug) cout << " edLst2 " << *edLst2 << endl;

   v1=0;                         // Total demand on left
   vt1=edLst1->getDemand();      // Total demand
   v2=0;                         // Total demand on right for 2nd
   vt2=edLst2->getDemand();      // Total demand for 2nd  

   v1 = edLst1->getDemandTo(fIndex+1,prob); // Total demand on left for first
   //if(myDebug) cout << " v1 " << v1 << endl;

   cc1 = edLst1->getDCost()+edLst2->getDCost();        // Total of parent costs
   totdemp = edLst1->getDemand()+edLst2->getDemand();  // Total of demands
   fm=edLst1->getNumberOfEdges();
   fn=edLst2->getNumberOfEdges();
   ccost = -1*cc1-1;

   for(m=fIndex;stay&&m<(edLst1->getNumberOfEdges()-1);m++) 
     if( (edge1 = edLst1->getNthEdge(m))!= NULL ) {
     fr1=edge1->getStartNode(); to1=edge1->getEndNode(); 
     v2 = edLst2->getDemandTo(sIndex+1,prob);
     for(n=sIndex;stay&&n<(edLst2->getNumberOfEdges()-1);n++) 
       if( (edge2 = edLst2->getNthEdge(n))!=NULL ) {
	  fr2 = edge2->getStartNode();
	  to2 = edge2->getEndNode();

	  retValue = ( fr1!=fr2 || to1!=to2 ) && 
	     (v1+vt2-v2)<=MVCap && (v2+vt1-v1)<=MVCap && 
	     edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
	     (edge1->getEndDateFirstNode()+edgeGraph->getTravelTime(fr1,to2))<=edge2->getStartDateSecondNode() &&
             edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) &&
	     (edge2->getEndDateFirstNode()+edgeGraph->getTravelTime(fr2,to1))<=edge1->getStartDateSecondNode() ;

          if(retValue) {
             tmp1 = defineELists(edLst1,m,edLst2,n,prob);
	     tmp2 = defineELists(edLst2,n,edLst1,m,prob);
             retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
	     if(retValue) { 
		cc2 = tmp1->getDCost()+tmp2->getDCost();
		switch(greedy) {
		  case 2:
		    totdemc = tmp1->getDemand()+tmp2->getDemand();
		    pdcostc = (totdemc>0?cc2/totdemc:FLT_MAX);
		    if( !found || (pdcostp-pdcostc)>ccost) { 
		      if( found ) { delete retE[0]; delete retE[1]; }
		      found=1; fm=m;fn=n;
		      ccost=pdcostp-pdcostc;
		      retE[0]=tmp1; retE[1]=tmp2; 
		    } else { delete tmp1; delete tmp2; }
		    break;
		  default:
		    if( !found || (cc1-cc2)>ccost ) {
		      if( found ) { delete retE[0]; delete retE[1]; }
		      found=1; fm=m;fn=n;
		      ccost=cc1-cc2; 
		      retE[0]=tmp1; retE[1]=tmp2; 
		    } else { delete tmp1; delete tmp2; }
		    break;
		}// end of switch()

	        if(fakeGreedy&&found&&ccost>0) stay=0;

	     } else { delete tmp1; delete tmp2; tmp1=tmp2=NULL; }

          }// end of if

	  if(prob->getCustomer(to2,cus)) v2+=cus.getDemand();

       }// end of if

     if(prob->getCustomer(to1,cus)) v1+=cus.getDemand();

   }// end of if

   if(found) { ret[0]=retE[0]; ret[1]=retE[1]; }

   rfIndex=fm; rsIndex=fn; 

   //if(myDebug) 
   cout << " Out of VRPGAOperation::find1PXG(EdgeListPtr) " << endl;

   return found;

}// end of find1PXG()

/*************************************************************************/
/*************************************************************************/
EdgeListPtr VRPGAOperation::defineELists( EdgeListPtr edLst1, int fIndex, 
			 EdgeListPtr edLst2, int sIndex, VRProblem* prob ) {
  return EdgeList::combineEdgeLists(edLst1,fIndex,edLst2,sIndex,prob);
}// end of defineELists()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
VRPGAOperation::find1PX( ListOfEdgeLists* el, ListOfEdgeLists* re, 
	                 ListOfEdgeLists** ret, VRProblem* prob) {
   int retValue = 0;
   EdgeGraph *edgeGraph = prob->getEdgeGraph();
   EdgeListPtr edLst1=NULL;
   EdgeListPtr edLst2=NULL;
   EdgeListPtr tmp1=NULL;
   EdgeListPtr tmp2=NULL;

   int i, j, m, n, iIndex, jIndex;

   int sizeEl = el->getNumberOfEdgeLists();
   int sizeRe = re->getNumberOfEdgeLists();
   iIndex = RandomGenerator::generateInt(0,sizeEl); i=0;

   //if(myDebug) cout << " In VRPGAOperation::find1PX( ListOfEdgeLists* el..) " << endl;

   while(i<el->getNumberOfEdgeLists() && !retValue) {
      edLst1 = el->getNthEdgeList((i+iIndex)%sizeEl);
      jIndex= RandomGenerator::generateInt(0,sizeRe); j=0;
      while(j<re->getNumberOfEdgeLists() && !retValue) {
	 edLst2 = re->getNthEdgeList((j+jIndex)%sizeRe);
	 if(find1PX(edLst1,1,edLst2,1,m,n,prob)) {
           tmp1=defineELists(edLst1,m,edLst2,n,prob);
           tmp2=defineELists(edLst2,n,edLst1,m,prob);
           retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
           if(!retValue) { delete tmp1; delete tmp2; tmp1=tmp2=NULL;}
	 }// end of if
	 if(!retValue) j++;
      }// end of while
      if(!retValue) i++;
   }// end of while

   if(retValue) {
     ret[0] = defineLOfELists2(el,(i+iIndex)%sizeEl,tmp1,prob);
     ret[1] = defineLOfELists2(re,(j+jIndex)%sizeRe,tmp2,prob);
   }// end of if

   //if(myDebug) cout << " Out of VRPGAOperation::find1PX( ListOfEdgeLists* el..) " << endl;

   return retValue;

}// end of find1PX()

/*************************************************************************/
/**                                                                      
   Applies Greedy Choice for each edgelist and finds the best possible
   point.
*/
/*************************************************************************/
VRPGAOperation::find1PXG( ListOfEdgeLists* el, ListOfEdgeLists* re, 
	                  ListOfEdgeLists** ret, VRProblem* prob, 
			  int greedy, int fakeGreedy) {

   EdgeGraph   *edgeGraph = prob->getEdgeGraph();
   EdgeListPtr edLst1=NULL;
   EdgeListPtr edLst2=NULL;
   EdgeListPtr tmp1=NULL;
   EdgeListPtr tmp2=NULL;
   EdgeListPtr retE[2];
   EdgeListPtr tmpE[2];

   int   i, j, m, n, k, l, iIndex, jIndex, found=0, fi, fj, stay=1, retValue=0;
   float ccost, cc1,cc2, totdemp, totdemc, pdcostp, pdcostc;

   int sizeEl = el->getNumberOfEdgeLists();
   int sizeRe = re->getNumberOfEdgeLists();
   iIndex = RandomGenerator::generateInt(0,sizeEl); 

   //if(myDebug) 
      cout << " In VRPGAOperation::find1PXG( ListOfEdgeLists* el..) " << endl;

   for(i=0;stay&&i<el->getNumberOfEdgeLists();i++) 
    if( (edLst1 = el->getNthEdgeList((i+iIndex)%sizeEl))!=NULL ) {
      jIndex= RandomGenerator::generateInt(0,sizeRe); j=0;
      for(j=0;stay&&j<re->getNumberOfEdgeLists();j++) 
       if( (edLst2 = re->getNthEdgeList((j+jIndex)%sizeRe))!=NULL &&
	   find1PXG(edLst1,1,edLst2,1,m,n,tmpE,greedy,prob,fakeGreedy) ) {

	   cc1     = edLst1->getDCost()+edLst2->getDCost();
  	   cc2     = tmpE[0]->getDCost()+tmpE[1]->getDCost();

  	   switch(greedy) {
	     case 2:
	      totdemp = edLst1->getDemand()+edLst2->getDemand();
	      pdcostp = (totdemp>0?cc1/totdemp:FLT_MAX);
	      totdemc = tmp1->getDemand()+tmp2->getDemand();
	      pdcostc = (totdemc>0?cc2/totdemc:FLT_MAX);
	      if( !found || (pdcostp-pdcostc)>ccost) { 
	        if( found ) { delete retE[0]; delete retE[1]; }
	        found=1; ccost=pdcostp-pdcostc;
	        retE[0]=tmpE[0];retE[1]=tmpE[1];
	        fi=(i+iIndex)%sizeEl; fj=(j+jIndex)%sizeRe;
	      } else { delete tmpE[0]; delete tmpE[1]; }
	      break;
	     default:
	      if( !found || (cc1-cc2)>ccost ) {
	        if( found ) { delete retE[0]; delete retE[1]; }
	        found=1; 
	        ccost=cc1-cc2; retE[0]=tmpE[0];retE[1]=tmpE[1];
	        fi=(i+iIndex)%sizeEl; fj=(j+jIndex)%sizeRe;
	      } else { delete tmpE[0]; delete tmpE[1]; }
	      break;
	   }// end of switch()

	   if(fakeGreedy&&found&&ccost>0) stay=0;

       }// end of if
    }// end of if

   cout << " found " << found << endl;
   if(found) {
     cout << " el is " << (el==NULL?"":"*NOT*") << " NULL! " << endl; 
     cout << " retE[0] is " << (retE[0]==NULL?"":"*NOT*") << " NULL! " << endl; 
     ret[0] = defineLOfELists2(el,fi,retE[0],prob);
     cout << " re is " << (re==NULL?"":"*NOT*") << " NULL! " << endl; 
     cout << " retE[1] is " << (retE[1]==NULL?"":"*NOT*") << " NULL! " << endl; 
     ret[1] = defineLOfELists2(re,fj,retE[1],prob);
   }// end of if

   //if(myDebug) 
      cout<<" Out of VRPGAOperation::find1PXG( ListOfEdgeLists* el..) "<<endl;

   return found;

}// end of find1PXG()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
VRPGAOperation::find1PX( ListOfEdgeLists* el, int fIndex,
			 ListOfEdgeLists* re, int sIndex,
			 int& rfIndex, int& rsIndex,
	                 ListOfEdgeLists** ret, VRProblem* prob) {
   int retValue = 0;
   EdgeGraph *edgeGraph = prob->getEdgeGraph();
   EdgeListPtr edLst1=NULL;
   EdgeListPtr edLst2=NULL;
   EdgeListPtr tmp1=NULL;
   EdgeListPtr tmp2=NULL;

   int i, j, m, n;

   int sizeEl = el->getNumberOfEdgeLists();
   int sizeRe = re->getNumberOfEdgeLists();

   if(myDebug) cout << " In VRPGAOperation::find1PX( ListOfEdgeLists* el, int fIndex,..) " << endl;

   i=fIndex;
   while(i<sizeEl && !retValue) {
      edLst1 = el->getNthEdgeList(i%sizeEl); j=sIndex;
      while(j<sizeRe && !retValue) {
	 edLst2 = re->getNthEdgeList(j%sizeRe);
	 if(find1PX(edLst1,1,edLst2,1,m,n,prob)) {
           tmp1=defineELists(edLst1,m,edLst2,n,prob);
           tmp2=defineELists(edLst2,n,edLst1,m,prob);
           retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
           if(!retValue) { delete tmp1; delete tmp2; tmp1=tmp2=NULL;}
	 }// end of if
	 if(!retValue) j++;
      }// end of while
      if(!retValue) i++;
   }// end of while

   rfIndex=i; rsIndex=j;
   if(retValue) {
     ret[0] = defineLOfELists2(el,i%sizeEl,tmp1,prob);
     ret[1] = defineLOfELists2(re,j%sizeRe,tmp2,prob);
   }// end of if

   if(myDebug) cout << " Out of VRPGAOperation::find1PX( ListOfEdgeLists* el, int fIndex,..) " << endl;
   return retValue;

}// end of find1PX()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
VRPGAOperation::find2PX( EdgeListPtr eList1, EdgeListPtr eList2, int& f0, 
			 int& f1, int& s0, int& s1, VRProblemPtr prob) {
  return find2PX(eList1,1,eList2,1,f0,f1,s0,s1,prob);
}// end of find2PX()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
VRPGAOperation::find2PX( EdgeListPtr eList1, int fs, 
			 EdgeListPtr eList2, int ss,
			 int& f0,int& f1,int& s0,int& s1, VRProblemPtr prob) {
   EdgeListPtr retE[2];
   int retValue=find2PX(eList1,fs,eList2,ss,f0,f1,s0,s1,prob,retE);
   if(retValue) { delete retE[0]; delete retE[1]; }
   return retValue;
}// end of find2PX()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
VRPGAOperation::find2PX( EdgeListPtr eList1, int fs, EdgeListPtr eList2, int ss,
			 int& f0, int& f1, int& s0, int& s1, 
			 VRProblemPtr prob, EdgeListPtr* retE) {
   int i, j, m, n, retValue = 0;
   EdgeListPtr tmp1 = NULL;
   EdgeListPtr tmp2 = NULL;
   ListOfEdgeLists* reti[2];
   if(myDebug) 
    cout << " in VRPGAOperation::find2PX(EdgeListPtr,..) " << endl;
   int st1 = maximum(fs,1); 
   int st2 = maximum(ss,1); 
   vrpcapacity MVCap=prob->getMaximumCapacity();
   vrpcapacity vt1=eList1->getDemand();
   vrpcapacity vt2=eList2->getDemand();
   vrpcapacity v11,v12, v21,v22;
   if(find1PX(eList1,st1,eList2,st2,i,j,prob)) {
     int fr1,to1,fr2,to2;
     EdgeGraph* edgeGraph = prob->getEdgeGraph();
     EdgePtr edge1, edge2;
     v11 = eList1->getDemandTo(i+1,prob);
     v21 = eList2->getDemandTo(j+1,prob);
     m=i+1;
     while(m<eList1->getNumberOfEdges()&&!retValue) {
       if( (edge1=eList1->getNthEdge(m))!=NULL ) {
	 fr1 = edge1->getStartNode();
	 to1 = edge1->getEndNode();
         n=j+1;
         while(n<eList2->getNumberOfEdges()&&!retValue) {
           if( (edge2=eList2->getNthEdge(n))!=NULL ) {
	     fr2 = edge2->getStartNode();
	     to2 = edge2->getEndNode();

	     retValue = ( fr1!=fr2 || to1!=to2 ) && 
	       //edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
               //edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) ;
	     edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
	     (edge1->getEndDateFirstNode()+edgeGraph->getTravelTime(fr1,to2))<=edge2->getStartDateSecondNode() &&
             edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) &&
	     (edge2->getEndDateFirstNode()+edgeGraph->getTravelTime(fr2,to1))<=edge1->getStartDateSecondNode() ;
             if(retValue) {
		v12 = eList1->getDemandTo(m+1,prob); 
		v22 = eList2->getDemandTo(n+1,prob); 
		retValue = ((v11+(v22-v21)+(vt1-v12))<=MVCap)&&
		           ((v21+(v12-v11)+(vt2-v22))<=MVCap);
	     }// end of if 
             if(retValue) {
	      tmp1 = defineELists(eList2,j,n,eList1,i,m,prob);
              tmp2 = defineELists(eList1,i,m,eList2,j,n,prob);
              retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
	      if(!retValue) { delete tmp1; delete tmp2; }
             }// end of if
	   }// end of if
	   if(!retValue) n++;
         }// end of while
       }// end of if
       if(!retValue) m++;
     }// end of while

     f0=i; f1=m; s0=j; s1=n; 
     if(retValue) { retE[0]=tmp1; retE[1]=tmp2; }

   }// end of if
   if(myDebug) 
     cout << " out of VRPGAOperation::find2PX(EdgeListPtr,..) " << endl;
   return retValue;

}// end of find2PX()

/*************************************************************************/
/*************************************************************************/
VRPGAOperation::satisfies2PX(VRPSolution *child, ListOfEdgeLists *el,
		             ListOfEdgeLists *re, ListOfEdgeLists** ret) {
   int retValue = 0;
   EdgeGraph *edgeGraph = child->getEdgeGraph();
   VRProblemPtr prob = child->getProblem();
   EdgeListPtr edLst1;
   EdgeListPtr edLst2;
   int i, j, m, n, fr1, to1, fr2, to2, k, l, iIndex, jIndex, f0, f1, s0, s1;

   int sizeEl = el->getNumberOfEdgeLists();
   int sizeRe = re->getNumberOfEdgeLists();
   iIndex = RandomGenerator::generateInt(0,sizeEl);
   i=0;

   if(myDebug) cout << "in VRPGAOperation::satisfies2PX() " << endl; 
   while(i<el->getNumberOfEdgeLists() && !retValue) {
      edLst1 = el->getNthEdgeList((i+iIndex)%sizeEl);
      j=0; jIndex= RandomGenerator::generateInt(0,sizeRe);
      while(j<re->getNumberOfEdgeLists() && !retValue) {
	 edLst2 = re->getNthEdgeList((j+jIndex)%sizeRe);
	 if(find2PX(edLst1,edLst2,f0,f1,s0,s1,prob)) {
	    EdgeListPtr tmp1 = defineELists(edLst1,f0,f1,edLst2,s0,s1,prob);
	    EdgeListPtr tmp2 = defineELists(edLst2,s0,s1,edLst1,f0,f1,prob);
	    retValue = tmp1->isValid(prob) && tmp2->isValid(prob);
	    if(!retValue) { delete tmp1; delete tmp2; }
	    else { 
	       ret[0]=defineLOfELists2(el,(i+iIndex)%sizeEl,tmp1,prob);
	       ret[1]=defineLOfELists2(re,(j+jIndex)%sizeRe,tmp2,prob);
            }// end of if
	 }// end of if
	 if(!retValue) j++;
      }// end of while
      if(!retValue) i++;
   }// end of while
    
   if(myDebug) cout << "out of VRPGAOperation::satisfies2PX() " << endl; 
   return retValue;

}// end of satisfies2PX()

/*************************************************************************/
/**                                                                      
    Based on Greedy Choice finds the best possible 2PX combination.
*/
/*************************************************************************/
VRPGAOperation::find2PXG( EdgeListPtr eList1, EdgeListPtr eList2, int& f0,
		  int& f1,int& s0,int& s1, VRProblemPtr prob, int greedy,
		  EdgeListPtr* retE, int fakeGreedy) {
   int i, j, m, n, retValue=0, ri, rj, rn, rm, stay=1, st1, st2, found=0 ; 
   float ccost, cc1,cc2, totdemp, totdemc, pdcostp, pdcostc;
   vrpcapacity MVCap=prob->getMaximumCapacity();
   vrpcapacity vt1=eList1->getDemand();
   vrpcapacity vt2=eList2->getDemand();
   vrpcapacity v11,v12, v21,v22;

   if(myDebug) cout << " in VRPGAOperation::find2PXG(EdgeListPtr,..) " << endl;

   cc1 = eList1->getDCost()+eList2->getDCost();
   totdemp = eList1->getDemand()+eList2->getDemand();
   pdcostp = (totdemp>0?cc1/totdemp:FLT_MAX);

   for(st1=1;stay&&st1<eList1->getNumberOfEdges()-1;st1++) 
     for(st2=1;stay&&st2<eList2->getNumberOfEdges()-1;st2++) 
       if(find1PX(eList1,st1,eList2,st2,i,j,prob)) {
         int fr1,to1,fr2,to2;
         EdgeGraph* edgeGraph = prob->getEdgeGraph();
         EdgePtr edge1, edge2;
         v11 = eList1->getDemandTo(i+1,prob);
         v21 = eList2->getDemandTo(j+1,prob);
         m=i+1;
         while(stay&&m<eList1->getNumberOfEdges()) {
           if( (edge1=eList1->getNthEdge(m))!=NULL ) {
	     fr1 = edge1->getStartNode();
	     to1 = edge1->getEndNode();
             n=j+1;
             while(stay&&n<eList2->getNumberOfEdges()) {
               if( (edge2=eList2->getNthEdge(n))!=NULL ) {
	         fr2 = edge2->getStartNode();
	         to2 = edge2->getEndNode();

	         retValue = ( fr1!=fr2 || to1!=to2 ) && 
	           //edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
                   //edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) ;
	           edgeGraph->doesEdgeExist(fr1,to2,edge1->getStartDate()) &&
	           (edge1->getEndDateFirstNode()+edgeGraph->getTravelTime(fr1,to2))<=edge2->getStartDateSecondNode() &&
                   edgeGraph->doesEdgeExist(fr2,to1,edge2->getStartDate()) &&
	           (edge2->getEndDateFirstNode()+edgeGraph->getTravelTime(fr2,to1))<=edge1->getStartDateSecondNode() ;
                 if(retValue) {
		   v12 = eList1->getDemandTo(m+1,prob); 
		   v22 = eList2->getDemandTo(n+1,prob); 
		   retValue = ((v11+(v22-v21)+(vt1-v12))<=MVCap)&&
		              ((v21+(v12-v11)+(vt2-v22))<=MVCap);
	         }// end of if 

                 if(retValue) {
	           EdgeListPtr tmp1 = defineELists(eList2,j,n,eList1,i,m,prob);
                   EdgeListPtr tmp2 = defineELists(eList1,i,m,eList2,j,n,prob);
                   retValue = (tmp1->isValid(prob) && tmp2->isValid(prob));
		   if(retValue) {

		     cc2 = tmp1->getDCost()+tmp2->getDCost();
		     switch(greedy) {
		      case 2:
                       totdemc = tmp1->getDemand()+tmp2->getDemand();
                       pdcostc = (totdemc>0?cc2/totdemc:FLT_MAX);
		       if( !found || (pdcostp-pdcostc)>ccost) { 
			  if(found) { delete retE[0]; delete retE[1]; }
		          found=1; ccost=pdcostp-pdcostc; ri=i;rj=j;rn=n;rm=m;
			  retE[0]=tmp1; retE[1]=tmp2;
		       } else { delete tmp1; delete tmp2; }
		       break;
		      default:
		       if( !found || (cc1-cc2)>ccost ) {
			 if(found) { delete retE[0]; delete retE[1]; }
		         found=1; ccost=cc1-cc2; ri=i;rj=j;rn=n;rm=m;
			 retE[0]=tmp1; retE[1]=tmp2;
		       } else { delete tmp1; delete tmp2; }
		       break;
                     }// end of switch()

		     if(fakeGreedy&&found&&ccost>0) stay=0;

                   } else { delete tmp1; delete tmp2; }

                 }// end of if
	       }// end of if
	       n++;
             }// end of while
           }// end of if
           m++;
         }// end of while

	 st2=j;

         //if(found) { f0=i; f1=m; s0=j;s1=n; }

       } else st2=eList2->getNumberOfEdges(); // end of if
   if(myDebug) cout << " out of VRPGAOperation::find2PXG(EdgeListPtr,..) " << endl;
   if(found) { f0=ri; f1=rm; s0=rj;s1=rn; }
   return found;
}// end of find2PXG()

/*************************************************************************/
/*************************************************************************/
VRPGAOperation::find2PXG( EdgeListPtr eList1, EdgeListPtr eList2, int& f0,
		  int& f1,int& s0,int& s1, VRProblemPtr prob, int greedy,
		  int fakeGreedy) {
  EdgeListPtr ret[2];
  int retValue = find2PXG(eList1,eList2,f0,f1,s0,s1,prob,greedy,ret,fakeGreedy);
  if(retValue) { delete ret[0]; delete ret[1];}
  return retValue;
}// end of find2PXG()

/*************************************************************************/
/*************************************************************************/
VRPGAOperation::satisfies2PXG( VRPSolution *child, ListOfEdgeLists *el,
		               ListOfEdgeLists *re, ListOfEdgeLists** ret,
			       int greedy, int fakeGreedy) {
   int retValue = 0;
   EdgeGraph *edgeGraph = child->getEdgeGraph();
   VRProblemPtr prob = child->getProblem();
   EdgeListPtr edLst1;
   EdgeListPtr edLst2;
   EdgeListPtr tmp[2];
   ListOfEdgeLists* tmp1;
   ListOfEdgeLists* tmp2;
   int i, j, m, n, fr1, to1, fr2, to2, k, l, iIndex, jIndex;
   int stay=1, f0, f1,s0, s1, found=0;
   float ccost, cc1,cc2, totdemp, totdemc, pdcostp, pdcostc;

   int sizeEl = el->getNumberOfEdgeLists();
   int sizeRe = re->getNumberOfEdgeLists();
   iIndex = RandomGenerator::generateInt(0,sizeEl);

   cc1 = el->getDCost()+re->getDCost();
   totdemp = el->getDemand()+re->getDemand();
   pdcostp = (totdemp>0?cc1/totdemp:FLT_MAX);

   if(myDebug) cout << "in VRPGAOperation::satisfies2PX() " << endl; 
   for(i=0;stay&&i<el->getNumberOfEdgeLists();i++) {
      edLst1 = el->getNthEdgeList((i+iIndex)%sizeEl);
      jIndex= RandomGenerator::generateInt(0,sizeRe);
      for(j=0;stay&&j<re->getNumberOfEdgeLists();j++) {
	 edLst2 = re->getNthEdgeList((j+jIndex)%sizeRe);
	 if(find2PXG(edLst1,edLst2,f0,f1,s0,s1,prob,greedy,tmp,fakeGreedy)) {
	    tmp1 = defineLOfELists2(el,(i+iIndex)%sizeEl,tmp[0],prob);
	    tmp2 = defineLOfELists2(re,(j+jIndex)%sizeRe,tmp[1],prob);
	    cc2 = tmp1->getDCost()+tmp2->getDCost();
	    switch(greedy) {
	      case 2:
               totdemc = tmp1->getDemand()+tmp2->getDemand();
               pdcostc = (totdemc>0?cc2/totdemc:FLT_MAX);
               if( !found || (pdcostp-pdcostc)>ccost) { 
		 if(found) { delete ret[0]; delete ret[1]; }
	         found=1; ccost=pdcostp-pdcostc; 
	         ret[0]=tmp1; ret[1]=tmp2;
	       } else { delete tmp1; delete tmp2; } 
	       break;
	      default:
	       if( !found || (cc1-cc2)>ccost ) {
		 if(found) { delete ret[0]; delete ret[1]; }
	         found=1; ccost=cc1-cc2; 
	         ret[0]=tmp1; ret[1]=tmp2;
	       } else { delete tmp1; delete tmp2; } 
	       break;
            }// end of switch()
	    if(fakeGreedy&&found&&ccost>0) stay=0;
	 }// end of if
      }// end of for
   }// end of for
    
   if(myDebug) cout << "out of VRPGAOperation::satisfies2PXG() " << endl; 
   return found;

}// end of satisfies2PXG()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists2( ListOfEdgeLists* el,int i, 
		                 EdgeListPtr temp, VRProblemPtr prob) {

   ListOfEdgeLists* ret = new ListOfEdgeLists();
   int k=0;
   if(el==NULL) { 
      cout << "In VRPGAOperation::defineLOfELists2() el is NULL!" << endl;
      exit(1);
   }// end of if
   if(prob==NULL) { 
      cout << "In VRPGAOperation::defineLOfELists2() prob is NULL!" << endl;
      exit(1);
   }// end of if
   while(k<i&&k<el->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k)),prob); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }
   cout << " temp will be added ! i:" << i << endl;
   if(temp!=NULL) ret->addEdgeList(temp,prob);
   cout << " temp is added " << endl;
   k=i+1;
   cout << " rest will be added el->getNumberOfEdgeLists(): " 
	<< el->getNumberOfEdgeLists() << endl;
   while(k<el->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k))); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }
   cout << "Out of VRPGAOperation::defineLOfELists2() " << endl;
   return ret;

}// end of defineLOfELists2()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists( ListOfEdgeLists* el,int i, 
		                 EdgeListPtr temp, VRProblemPtr prob) {

   ListOfEdgeLists* ret = new ListOfEdgeLists();
   int k=0;
   while(k<i) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k)),prob); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }
   //if(temp!=NULL) ret->addEdgeList(new EdgeList(*temp),prob);
   if(temp!=NULL) ret->addEdgeList(temp,prob);
   k=i+1;
   while(k<el->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k))); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }
   return ret;

}// end of defineLOfELists()

/*************************************************************************/
/*************************************************************************/
EdgeListPtr VRPGAOperation::defineELists( EdgeListPtr first, int f0, int f1,
			 EdgeListPtr second, int s0, int s1, VRProblem* prob ) {
  return EdgeList::combineEdgeLists(first,f0,f1,second,s0,s1,prob);
}// end of defineELists()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists( ListOfEdgeLists* el,int f0, int f1, 
                                 ListOfEdgeLists* re,int s0, int s1,
		                 VRProblemPtr prob) {
  return defineLOfELists(el,f0,f1,re,s0,s1,NULL,NULL,prob);
}// end of defineLOfELists()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists( ListOfEdgeLists* el,int f0, 
				 ListOfEdgeLists* re,int s0, 
		                 VRProblemPtr prob) {
   return defineLOfELists(el,f0,NULL,re,s0,prob);
}// end of defineLOfELists()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists( ListOfEdgeLists* el,int f0, int f1, 
                                 ListOfEdgeLists* re,int s0, int s1,
				 EdgeListPtr e1, EdgeListPtr e2, 
				 VRProblemPtr prob) {
   ListOfEdgeLists* ret = new ListOfEdgeLists();
   int k=0;

   while(k<f0&&k<el->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k)),prob); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }

   //if(e1!=NULL) ret->addEdgeList(new EdgeList(*e1));
   if(e1!=NULL) ret->addEdgeList(e1,prob);

   k=s0+1;
   while(k<s1&&k<re->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*re->getNthEdgeList(k)),prob); k++; }
   { ret->addEdgeList(re->getNthEdgeList(k),prob); k++; }

   //if(e2!=NULL) ret->addEdgeList(new EdgeList(*e2));
   if(e2!=NULL) ret->addEdgeList(e2,prob);

   k=f1+1;
   while(k<el->getNumberOfEdgeLists()) 
   //{ ret->addEdgeList(new EdgeList(*el->getNthEdgeList(k)),prob); k++; }
   { ret->addEdgeList(el->getNthEdgeList(k),prob); k++; }
   
   return ret;

}// end of defineLOfELists()

/*************************************************************************/
/**                                                                      */
/*************************************************************************/
ListOfEdgeLists* 
VRPGAOperation::defineLOfELists( ListOfEdgeLists* el,int f0, 
                                 EdgeListPtr e1,
				 ListOfEdgeLists* re,int s0, 
		                 VRProblemPtr prob) {
   return defineLOfELists( el, f0, el->getNumberOfEdgeLists(),
			   re, s0, re->getNumberOfEdgeLists(), e1, NULL, prob);
}// end of defineLOfELists()

/*************************************************************************/
/** */
/**
   1) Copy all the ListOfEdgeLists from parent to child as long as
      they do not have any common nodes with the "el".
   2) Add "el" to the "child".
   3) call patchUp()
*/
/*************************************************************************/
void VRPGAOperation::buildChildren(VRPChromosomes* child,
		  VRPChromosomes* parent, ListOfEdgeLists* el) {
  int i;
  ListOfEdgeLists* edLst ;

  child->cleanUpVehicles();
  child->addVehicleSchedule(*el);

  for(i=0;i<parent->getNumOfRotations();i++)
    if( (edLst=parent->getListOfEdgeListsIndex(i))!=NULL )
      child->updateVehicles(edLst);

  child->patchUp();

}// end of buildChildren()

/*************************************************************************/
/**
*/
/**
   1) Copy all the ListOfEdgeLists from parent to child as long as
      they do not have any common nodes with the "el".
   2) Add "el" to the "child".
   3) call patchUp()
*/
/*************************************************************************/
void VRPGAOperation::buildChildren(VRPSolution* child,VRPSolution* parent,
				int n0, int n1, ListOfEdgeLists** el) {
  int i;
  ListOfEdgeLists* edLst ;

  child->cleanUpVehicles();

  child->addVehicleSchedule(*el[0]);
  child->addVehicleSchedule(*el[1]);

  for(i=0;i<parent->getNumOfRotations();i++)
   if(i!=n0&&i!=n1) {
     edLst = parent->getListOfEdgeListsIndex(i);
     if( edLst!=NULL ) child->updateVehicles(edLst);
   }// end of if

  child->patchUp();

}// end of buildChildren()

/*************************** T H E   E N D *******************************/
